;
; Extended memory driver for the CSORY 256K memory expansion.
;
; Marco van den Heuvel, 2010-01-29
;

	.include 	"zeropage.inc"

	.include	"em-kernel.inc"
	.include	"em-error.inc"
	.include	"plus4.inc"

	.macpack	generic


; ------------------------------------------------------------------------
; Header. Includes jump table

.segment	"JUMPTABLE"

; Driver signature

	.byte	$65, $6d, $64           ; "emd"
	.byte	EMD_API_VERSION		; EM API version number

; Jump table.

	.word	INSTALL
	.word	UNINSTALL
	.word	PAGECOUNT
	.word	MAP
	.word	USE
	.word	COMMIT
	.word	COPYFROM
	.word	COPYTO

; ------------------------------------------------------------------------
; Constants

BASE		= $8000
EXP_SELECT	= $FD16
CSORY_CTRL	= $FD15
PAGES		= 256 * 4

; ------------------------------------------------------------------------
; Data.

.bss
curpage:        .res    2               ; Current page number
window:         .res    256             ; Memory "window"

.code

; ------------------------------------------------------------------------
; INSTALL routine. Is called after the driver is loaded into memory. If
; possible, check if the hardware is present and determine the amount of
; memory available.
; Must return an EM_ERR_xx code in a/x.
;

INSTALL:

	sei
	sta	ENABLE_RAM
	lda	#$20
	sta	EXP_SELECT
	lda	#$80
	sta	CSORY_CTRL
	ldx	BASE
	cpx	BASE
	bne	@notpresent
	lda	#$81
	sta	CSORY_CTRL
	stx	BASE
	inc	BASE
	cpx	BASE
	beq	@notpresent
	lda	#$80
	sta	CSORY_CTRL
	cpx	BASE
	bne	@notpresent
	sta	ENABLE_ROM
	lda	#<EM_ERR_OK
	ldx	#>EM_ERR_OK
	cli
	rts

@notpresent:
	sta	ENABLE_ROM
	ldx	#>EM_ERR_NO_DEVICE
	lda	#<EM_ERR_NO_DEVICE
	cli
;	rts                             ; Run into UNINSTALL instead

; ------------------------------------------------------------------------
; UNINSTALL routine. Is called before the driver is removed from memory.
; Can do cleanup or whatever. Must not return anything.
;

UNINSTALL:
	rts


; ------------------------------------------------------------------------
; PAGECOUNT: Return the total number of available pages in a/x.
;

PAGECOUNT:
	lda	#<PAGES
	ldx	#>PAGES
	rts

; ------------------------------------------------------------------------
; MAP: Map the page in a/x into memory and return a pointer to the page in
; a/x. The contents of the currently mapped page (if any) may be discarded
; by the driver.
;

MAP:
	sta	curpage			; Remember the new page
	stx	curpage+1
	jsr	adjust_bank_and_page
	stx	CSORY_CTRL
	clc
	adc	#>BASE
	sta	ptr1+1
	ldy	#0
	sty	ptr1

	lda	#<window
	sta	ptr2
	lda	#>window
	sta	ptr2+1

; Transfer one page

	jsr	transfer                ; Transfer one page

; Return the memory window

	lda	#<window
	ldx	#>window                ; Return the window address
	rts

; ------------------------------------------------------------------------
; USE: Tell the driver that the window is now associated with a given page.

USE:
	sta	curpage                 ; Remember the page
	stx	curpage+1
	lda	#<window
	ldx	#>window                ; Return the window
	rts

; ------------------------------------------------------------------------
; COMMIT: Commit changes in the memory window to extended storage.

COMMIT:
	lda	curpage			; Get the current page
	ldx	curpage+1
	jsr	adjust_bank_and_page
	stx	CSORY_CTRL
	clc
	adc	#>BASE
	sta	ptr2+1
	ldy	#0
	sty	ptr2

	lda	#<window
	sta	ptr1
	lda	#>window
	sta	ptr1+1

; Transfer one page. Y must be zero on entry. Because we bank out the
; kernal, we will run the routine with interrupts disabled but leave
; short breath times. Unroll the following loop to make it somewhat faster.

transfer:
	sei
	sta	ENABLE_RAM

@L1:
	lda	(ptr1),y
	sta	(ptr2),y
	iny
	bne	@L1
	sta	ENABLE_ROM
	cli
	rts

; ------------------------------------------------------------------------
; COPYFROM: Copy from extended into linear memory. A pointer to a structure
; describing the request is passed in a/x.
; The function must not return anything.
;

COPYFROM:
	sei
	jsr	setup

; Setup is:
;
;   - ptr1 contains the struct pointer
;   - ptr2 contains the linear memory buffer
;   - ptr3 contains -(count-1)
;   - ptr4 contains the page memory buffer
;   - tmp1 contains the bank value
;   - tmp2 contains the page memory buffer offset
;   - tmp3 contains zero (used for linear memory buffer offset)

	lda	#$20
	sta	EXP_SELECT
	lda	tmp1
	sta	CSORY_CTRL
	jmp	@L3

@L1:
	sta	ENABLE_RAM
	ldy	tmp2
	lda	(ptr4),y
	ldy	tmp3
	sta	(ptr2),y
	sta	ENABLE_ROM
	inc	tmp3
	bne	@L2
	inc	ptr2+1
@L2:
	inc	tmp2
	beq	@L4


; Bump count and repeat

@L3:
	inc	ptr3
	bne	@L1
	inc	ptr3+1
	bne	@L1
	cli
	rts

; Bump page register

@L4:
	inc	ptr4+1
	lda	ptr4+1
	cmp	#$c0
	bne	@L3
	lda	#>BASE
	sta	ptr4+1
	inc	tmp1
	lda	tmp1
	sta	CSORY_CTRL
	jmp	@L3


; ------------------------------------------------------------------------
; COPYTO: Copy from linear into extended memory. A pointer to a structure
; describing the request is passed in a/x.
; The function must not return anything.
;

COPYTO:
	sei
	jsr	setup

; Setup is:
;
;   - ptr1 contains the struct pointer
;   - ptr2 contains the linear memory buffer
;   - ptr3 contains -(count-1)
;   - ptr4 contains the page memory buffer
;   - tmp1 contains the bank value
;   - tmp2 contains the page memory buffer offset
;   - tmp3 contains zero (used for linear memory buffer offset)

	lda	#$20
	sta	EXP_SELECT
	lda	tmp1
	sta	CSORY_CTRL
	jmp	@L3

@L1:
	sta	ENABLE_RAM
	ldy	tmp3
	lda	(ptr2),y
	ldy	tmp2
	sta	(ptr4),y
	sta	ENABLE_ROM
	inc	tmp3
	bne	@L2
	inc	ptr2+1
@L2:
	inc	tmp2
	beq	@L4


; Bump count and repeat

@L3:
	inc	ptr3
	bne	@L1
	inc	ptr3+1
	bne	@L1
	cli
	rts

; Bump page register

@L4:
	inc	ptr4+1
	lda	ptr4+1
	cmp	#$c0
	bne	@L3
	lda	#>BASE
	sta	ptr4+1
	inc	tmp1
	lda	tmp1
	sta	CSORY_CTRL
	jmp	@L3


; ------------------------------------------------------------------------
; Helper function for COPYFROM and COPYTO: Store the pointer to the request
; structure and prepare data for the copy

setup:
	sta	ptr1
	stx	ptr1+1          ; Save passed pointer

; Get the page number from the struct and adjust it so that it may be used
; with the hardware. That is: ptr4 has the page address and page offset
; tmp1 will hold the bank value

	ldy	#EM_COPY::PAGE+1
	lda	(ptr1),y
	tax
	dey
	lda	(ptr1),y
	jsr	adjust_bank_and_page
	clc
	adc	#>BASE
	sta	ptr4+1
	stx	tmp1

; Get the buffer pointer into ptr2

	ldy	#EM_COPY::BUF
	lda	(ptr1),y
	sta	ptr2
	iny
	lda	(ptr1),y
	sta	ptr2+1

; Get the count, calculate -(count-1) and store it into ptr3

	ldy	#EM_COPY::COUNT
	lda	(ptr1),y
	eor	#$FF
	sta	ptr3
	iny
	lda	(ptr1),y
	eor	#$FF
	sta	ptr3+1

; Get the page offset into tmp2, clear tmp3 and clear ptr4 low byte

	ldy	#EM_COPY::OFFS
	lda	(ptr1),y
	sta	tmp2
	ldy	#$00
	sty	ptr4
	sty	tmp3

; Done

	rts


; Helper routine to correct for the bank and page

adjust_bank_and_page:
	sta	tmp4
	lda	#$80
	sta	tmp3
	lda	tmp4
	lsr
	lsr
	lsr
	lsr
	lsr
	lsr
	ora	tmp3
	sta	tmp3
	txa
	asl
	asl
	ora	tmp3
	sta	tmp3
	lda	tmp4
	and	#$3f
	ldx	tmp3
	rts
